# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Properties report writer for Seal5."""

import argparse
import logging
import pathlib

import pandas as pd

from m2isar.metamodel import arch

from seal5.model import Seal5OperandAttribute, Seal5InstrAttribute
from seal5.model_utils import load_model

logger = logging.getLogger("properties_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", nargs="+", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md"], default="auto")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    assert args.output is not None
    out_path = pathlib.Path(args.output)

    # resolve model paths
    properties_data = []
    top_levels = args.top_level
    for top_level in top_levels:
        top_level = pathlib.Path(top_level)

        model_obj = load_model(top_level, compat=args.compat)

        # preprocess model
        # print("model", model)
        for set_name, set_def in model_obj.sets.items():
            # print("set_name", set_name)
            xlen = set_def.xlen
            model = top_level.stem

            def get_set_properties(set_name, xlen):
                is_rv32 = xlen == 32
                is_rv64 = xlen == 64
                return {
                    "set": set_name,
                    "xlen": xlen,
                    "is_rv32": is_rv32,
                    "is_rv64": is_rv64,
                }

            for instr_def in set_def.instructions.values():
                instr_name = instr_def.name
                # print("instr_name", instr_name)
                operands = instr_def.operands

                def get_enc_properties(instr_def):
                    enc_size = instr_def.size
                    is_compressed = enc_size == 16
                    # TODO: opcode
                    # TODO: format

                    def detect_opcode(instr_def):  # TODO: move to transform and store as attr
                        OPCODE_LOOKUP = {
                            "LOAD": 0b00000,
                            "LOAD-FP": 0b00001,
                            "custom-0": 0b00010,
                            "MISC-MEM": 0b00011,
                            "OP-IMM": 0b00100,
                            "AUIPC": 0b00101,
                            "OP-IMM-32": 0b00110,
                            # "48bit": 0b00111,
                            "STORE": 0b01000,
                            "STORE-FP": 0b01001,
                            "custom-1": 0b01010,
                            "AMO": 0b01011,
                            "OP": 0b01100,
                            "LUI": 0b01101,
                            "OP-32": 0b01110,
                            # "64bit": 0b01111,
                            "MADD": 0b10000,
                            "MSUB": 0b10001,
                            "NMADD": 0b10010,
                            "NMSUB": 0b10011,
                            "OP-FP": 0b10100,
                            "OP-V": 0b10101,
                            "custom-2": 0b10110,  # rv128i
                            # "48bit2": 0b10111,
                            "BRANCH": 0b11000,
                            "JALR": 0b11001,
                            # "reserved": 0b11010,
                            "JAL": 0b11011,
                            "SYSTEM": 0b11100,
                            "OP-P": 0b11101,
                            "custom-3": 0b11110,
                            # "80bit+": 0b11111,
                        }
                        OPCODE_LOOKUP_REV = {v: k for k, v in OPCODE_LOOKUP.items()}
                        size = 0  # TODO: use instr_def.size
                        opcode = None
                        for e in reversed(instr_def.encoding):
                            # print("e", e, dir(e))
                            if isinstance(e, arch.BitVal):
                                length = e.length
                                if size == 0:
                                    val = e.value
                                    if length >= 7:
                                        val = val & 0b1111111
                                        opcode = val >> 2
                            elif isinstance(e, arch.BitField):
                                length = e.range.length
                            else:
                                assert False
                            size += length
                        if size != 32:  # TODO: support compressed opcodes
                            assert size == 16
                            return "COMPRESSED"
                        assert opcode is not None
                        found = OPCODE_LOOKUP_REV.get(opcode, None)
                        assert found is not None, f"Opcode not found: {bin(opcode)}"
                        return found

                    def detect_format(instr_def):  # TODO: move to transform and store as attr
                        # enc = instr_def.encoding
                        operands = instr_def.operands
                        char_lookup = {}
                        imm_char = "a"
                        reg_char = "A"
                        temp = ""
                        for e in reversed(instr_def.encoding):
                            if isinstance(e, arch.BitVal):
                                new = bin(e.value)[2:].zfill(e.length)
                                temp = f"{new}{temp}"
                            elif isinstance(e, arch.BitField):
                                length = e.range.length
                                found = char_lookup.get(e.name, None)
                                char = "#"
                                if found is not None:
                                    char = found
                                else:
                                    op = operands.get(e.name, None)
                                    if op is not None:
                                        if Seal5OperandAttribute.IS_REG in op.attributes:
                                            char = reg_char
                                            reg_char = chr(ord(reg_char) + 1)
                                        elif Seal5OperandAttribute.IS_IMM in op.attributes:
                                            char = imm_char
                                            imm_char = chr(ord(imm_char) + 1)
                                        else:
                                            char = "%"
                                    char_lookup[e.name] = char
                                new = char * length
                                temp = f"{new}{temp}"
                            else:
                                assert False
                        FMT_LOOKUP = [  # needs to be sorted by weight (number of non-?)
                            ("aaaaaaaBBBBBAAAAA???aaaaa???????", "s-type (mem)"),
                            ("??aaaaaCCCCCBBBBB???AAAAA???????", "f2-type (binop)"),
                            ("??bbbbbaaaaaBBBBB???AAAAA???????", "f2-type (unop)"),
                            ("???????CCCCCBBBBB???AAAAA???????", "r-type (binop)"),
                            ("???????BBBBBAAAAA???00000???????", "r-type (binop, noout)"),
                            ("???????00000BBBBB???AAAAA???????", "r-type (unop)"),
                            ("???????00000BBBBB???00000???????", "r-type (unop, noout)"),
                            ("???????0000000000???AAAAA???????", "r-type (nonop)"),
                            ("???????0000000000???00000???????", "r-type (nonop)"),  # s4emac
                            ("???????aaaaaBBBBB???AAAAA???????", "ri-type (binop)"),
                        ]
                        found2 = f"unknown[{temp}]"
                        for mask, fmt in FMT_LOOKUP:
                            if len(mask) != len(temp):
                                continue
                            masked = "".join([c if mask[i] != "?" else "?" for i, c in enumerate(temp)])
                            if masked == mask:
                                found2 = fmt
                                break
                        return found2

                    opcode = detect_opcode(instr_def)
                    enc_format = detect_format(instr_def)

                    return {
                        "enc_size": enc_size,
                        "is_compressed": is_compressed,
                        "enc_format": enc_format,
                        "opcode": opcode,
                    }

                def get_operands_properties(operands):
                    num_operands = len(operands)
                    num_inputs = len([op for op in operands.values() if Seal5OperandAttribute.IN in op.attributes])
                    num_outputs = len([op for op in operands.values() if Seal5OperandAttribute.OUT in op.attributes])
                    num_inouts = len([op for op in operands.values() if Seal5OperandAttribute.INOUT in op.attributes])
                    num_regs = len([op for op in operands.values() if Seal5OperandAttribute.IS_REG in op.attributes])
                    num_gprs = len(
                        [
                            op
                            for op in operands.values()
                            if Seal5OperandAttribute.IS_REG in op.attributes
                            and op.attributes[Seal5OperandAttribute.REG_CLASS] == "GPR"
                        ]
                    )
                    num_imms = len([op for op in operands.values() if Seal5OperandAttribute.IS_IMM in op.attributes])
                    imm_types = set(
                        op.attributes[Seal5OperandAttribute.TYPE]
                        for op in operands.values()
                        if Seal5OperandAttribute.IS_IMM in op.attributes
                    )
                    has_imm_leaf = any(
                        Seal5OperandAttribute.IS_IMM_LEAF in op.attributes
                        for op in operands.values()
                        if Seal5OperandAttribute.IS_IMM in op.attributes
                    )
                    is_multi_in = num_inputs + num_inouts > 1
                    is_single_in = num_inputs + num_inouts == 1
                    is_multi_out = num_outputs + num_inouts > 1
                    is_single_out = num_outputs + num_inouts == 1
                    is_mimo = is_multi_in and is_multi_out
                    is_miso = is_multi_in and is_single_out
                    is_siso = is_single_in and is_single_out
                    is_simo = is_single_in and is_multi_out
                    return {
                        "num_operands": num_operands,
                        "num_inputs": num_inputs,
                        "num_outputs": num_outputs,
                        "num_inouts": num_inouts,
                        "num_regs": num_regs,
                        "num_gprs": num_gprs,
                        "num_imms": num_imms,
                        "imm_types": imm_types,
                        "is_multi_in": is_multi_in,
                        "is_single_in": is_single_in,
                        "is_multi_out": is_multi_out,
                        "is_single_out": is_single_out,
                        "is_mimo": is_mimo,
                        "is_miso": is_miso,
                        "is_siso": is_siso,
                        "is_simo": is_simo,
                        "has_imm_leaf": has_imm_leaf,
                        # TODO: num_csrs, num_fprs, num_custom...
                    }

                def get_side_effect_properties(attrs):
                    uses_custom_reg = Seal5InstrAttribute.USES in attrs and len(attrs[Seal5InstrAttribute.USES]) > 0
                    defs_custom_reg = Seal5InstrAttribute.DEFS in attrs and len(attrs[Seal5InstrAttribute.DEFS]) > 0
                    return {
                        "has_side_effects": Seal5InstrAttribute.HAS_SIDE_EFFECTS in attrs,
                        "may_load": Seal5InstrAttribute.MAY_LOAD in attrs,
                        "may_store": Seal5InstrAttribute.MAY_STORE in attrs,
                        "is_terminator": Seal5InstrAttribute.IS_TERMINATOR in attrs,
                        "is_branch": Seal5InstrAttribute.IS_BRANCH in attrs,
                        "skip_pattern_gen": Seal5InstrAttribute.SKIP_PATTERN_GEN in attrs,
                        "uses_custom_reg": uses_custom_reg,
                        "defs_custom_reg": defs_custom_reg,
                    }

                attrs = instr_def.attributes
                data = {
                    "model": model,
                    **get_set_properties(set_name, xlen),
                    "instr": instr_name,
                    **get_enc_properties(instr_def),
                    **get_operands_properties(operands),
                    **get_side_effect_properties(attrs),
                    # TODO: operations
                }
                properties_data.append(data)
    properties_df = pd.DataFrame(properties_data)
    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        properties_df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        properties_df.to_pickle(out_path)
    elif fmt == "md":
        # properties_df.to_markdown(out_path, tablefmt="grid", index=False)
        properties_df.to_markdown(out_path, index=False)
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""PatternGen utils for seal5."""
from pathlib import Path
from typing import Optional

from seal5.logging import get_logger
from seal5.settings import PatchSettings
from seal5.types import PatchStage
from seal5.index import File, Directory, NamedPatch, write_index_yaml
from seal5.riscv_utils import build_riscv_mattr, get_riscv_defaults
from seal5 import utils

logger = get_logger()


def build_pattern_gen(
    src: Path,
    dest: Path,
    # debug: bool = False,
    use_ninja: bool = False,
    verbose: bool = False,
    cmake_options: Optional[dict] = None,
):
    """Build PatternGen tool."""
    if cmake_options is None:
        cmake_options = {}
    cmake_args = utils.get_cmake_args(cmake_options)
    dest.mkdir(exist_ok=True)
    utils.cmake(
        src / "llvm",
        *cmake_args,
        use_ninja=use_ninja,
        cwd=dest,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )
    utils.make(
        "pattern-gen", cwd=dest, print_func=logger.info if verbose else logger.debug, live=True, use_ninja=use_ninja
    )


def get_pattern_gen_patches(
    src: Path,
    temp_dir: Path,
    # verbose: bool = False,
):
    """Generate patches for PatternGen integration."""
    # TODO: copy!
    artifacts = []
    directory_artifact = Directory(
        "llvm/tools/pattern-gen",
        src_path=f"{src}/llvm/tools/pattern-gen",
    )
    artifacts.append(directory_artifact)
    file_artifact = File(
        "llvm/include/llvm/CodeGen/GlobalISel/PatternGen.h",
        src_path=f"{src}/llvm/include/llvm/CodeGen/GlobalISel/PatternGen.h",
    )
    artifacts.append(file_artifact)
    file_artifact = File(
        "llvm/lib/CodeGen/GlobalISel/PatternGen.cpp",
        src_path=f"{src}/llvm/lib/CodeGen/GlobalISel/PatternGen.cpp",
    )
    artifacts.append(file_artifact)
    patch_artifact = NamedPatch(
        "llvm/include/llvm/InitializePasses.h",
        key="initialize_passes_decl",
        content="void initializePatternGenPass(PassRegistry&);",
    )
    artifacts.append(patch_artifact)
    patch_artifact = NamedPatch(
        "llvm/lib/CodeGen/GlobalISel/CMakeLists.txt",
        key="gisel_cmake_srcs",
        content="  PatternGen.cpp",
    )
    artifacts.append(patch_artifact)
    patch_artifact = NamedPatch(
        "llvm/lib/CodeGen/GlobalISel/GlobalISel.cpp",
        key="gisel_init",
        content="  initializePatternGenPass(Registry);",
    )
    artifacts.append(patch_artifact)

    index_file = temp_dir / "pattern_gen_support_index.yml"
    write_index_yaml(index_file, artifacts, {}, content=True)
    patch_settings = PatchSettings(
        name="pattern_gen_support",
        stage=int(PatchStage.PHASE_0),
        comment="Integrate PatternGen in Seal5 LLVM",
        index=str(index_file),
        generated=True,
        target="llvm",
    )

    return patch_settings


def build_llc(
    src: Path,
    dest: Path,
    # debug: bool = False,
    use_ninja: bool = False,
    verbose: bool = False,
    cmake_options: Optional[dict] = None,
):
    """Build llc tool."""
    if cmake_options is None:
        cmake_options = {}
    cmake_args = utils.get_cmake_args(cmake_options)
    dest.mkdir(exist_ok=True)
    utils.cmake(
        src / "llvm",
        *cmake_args,
        use_ninja=use_ninja,
        cwd=dest,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )
    utils.make("llc", cwd=dest, print_func=logger.info if verbose else logger.debug, live=True, use_ninja=use_ninja)


def run_pattern_gen(
    build_dir: Path,
    src: Path,
    dest: Path,
    verbose: bool = False,
    ext=None,
    mattr=None,
    xlen=None,
    skip_formats=False,
    skip_patterns=False,
    skip_verify=True,
    no_extend=True,
    debug=False,
):
    """Excute pattern-gen executable."""
    if not isinstance(build_dir, Path):
        build_dir = Path(build_dir)
    pattern_gen_args = [src]
    # pattern_gen_args.extend(["-custom-legalizer-settings=foo", "-disable-gisel-legality-check"])

    if dest:
        pattern_gen_args.extend(["-o", str(dest)])

    if ext:
        # pattern_gen_args.extend(["--ext", ext])
        preds = [f"Has{ext}", f"IsRV{xlen}"]
        preds_str = ", ".join(preds)
        pattern_gen_args.extend(["-p", preds_str])

    if mattr is None:
        features, _ = get_riscv_defaults()
        if ext:
            ext_ = ext.lower()
            ext_ = ext_.replace("std", "").replace("vendor", "").replace("ext", "")
            features.append(f"+{ext_}")
        mattr = build_riscv_mattr(features, xlen=xlen)

    if mattr:
        pattern_gen_args.extend(["--mattr", mattr])

    if xlen:
        assert xlen in [32, 64]
        pattern_gen_args.extend(["--riscv-xlen", str(xlen)])

    if skip_patterns:
        pattern_gen_args.append("--skip-patterns")

    if skip_formats:
        pattern_gen_args.append("--skip-formats")

    if skip_verify:
        pattern_gen_args.append("--skip-verify")

    if debug:
        pattern_gen_args.append("--debug")

    if no_extend:
        pattern_gen_args.append("--no-extend")

    # break_on_err = True
    break_on_err = False

    # TODO: dump gmir?
    def handle_exit(code=None, out=None):
        if code is not None and code != 0:
            err_file = str(dest) + ".err"
            with open(err_file, "w", encoding="utf-8") as f:
                f.write(out)
        return code

    pattern_gen_exe = build_dir / "bin" / "pattern-gen"
    assert pattern_gen_exe.is_file(), "pattern-gen not found"

    try:
        out = utils.exec_getout(
            pattern_gen_exe,
            *pattern_gen_args,
            # cwd=dest,
            print_func=logger.info if verbose else logger.debug,
            handle_exit=handle_exit,
            live=True,
        )
    except Exception as e:
        if break_on_err:
            input("^^^ERR^^^")
        if dest.is_file():
            with open(dest, "r", encoding="utf-8") as f:
                content = f.read()
            if len(content) == 0:
                dest.unlink()
        raise e
    if not skip_patterns:
        # errs = None
        # opt_ll = None
        pat = []
        found_pattern = False
        # reason = None
        # rest = []
        is_err = False
        for line in out.split("\n"):
            # print("line", line)
            if found_pattern:
                # print("A1")
                pat.append(line)
                # found_pattern = False
            else:
                # print("A2")
                if "Pattern for" in line:
                    # print("B1")
                    pat = [line.split(":", 1)[1]]
                    # found_pattern = True
                elif "Pattern Generation failed for" in line:
                    # reason = line
                    is_err = True
        # print("pat", pat)
        if len(pat) > 0:
            pat = "\n".join(pat)
            pat_file = str(dest) + ".pat"
            with open(pat_file, "w", encoding="utf-8") as f:
                f.write(pat)
        else:
            is_err = True
        # else:
        #     if break_on_err:
        #         print("\n".join(rest))
        #         input("^^^Pattern not found^^^")
        if is_err:
            if break_on_err:
                print(out)
                input("^^^ERROR^^^")
            dest.unlink()
        out_file = str(dest) + (".err" if is_err else ".out")
        with open(out_file, "w", encoding="utf-8") as f:
            f.write(out)
        # if len(rest) > 0:
        #     rest = "\n".join(rest)
        #     rest_file = str(dest) + (".err" if is_err else ".out")
        #     with open(rest_file, "w") as f:
        #         f.write(rest)
        # if reason:
        #     reason_file = str(dest) + ".reason"
        #     with open(reason_file, "w") as f:
        #         f.write(reason)


def convert_ll_to_gmir(
    build_dir: Path,
    src: Path,
    dest: Path,
    mattr=None,
    xlen=None,
    optimize=3,
    verbose: bool = False,
):
    """Convert LLVM-IR file to GMIR file."""
    if mattr is None:
        features, _ = get_riscv_defaults()
        mattr = build_riscv_mattr(features, xlen=xlen)

    assert xlen is not None, "Needs XLEN"

    llc_args = [src, f"-mtriple=riscv{xlen}-unknown-elf", "-stop-after=irtranslator", "-global-isel", f"-O{optimize}"]

    if mattr:
        llc_args.extend(["--mattr", mattr])

    if not isinstance(build_dir, Path):
        build_dir = Path(build_dir)

    llc_exe = build_dir / "bin" / "llc"
    assert llc_exe.is_file(), "llc  not found"

    assert dest is not None
    llc_args.extend(["-o", str(dest)])

    _ = utils.exec_getout(
        llc_exe,
        *llc_args,
        # cwd=dest,
        print_func=logger.info if verbose else logger.debug,
        live=True,
    )



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel by removing unused constants."""

import sys
import argparse
import logging
import pathlib

import pandas as pd

from m2isar.metamodel import patch_model

import seal5.model
from seal5.model_utils import load_model, dump_model

from . import visitor
from .utils import IOMode

logger = logging.getLogger("detect_inouts")


class VisitorContext:
    def __init__(self):
        self.reads = set()
        self.writes = set()
        self.stack = []

    def push(self, mode):
        self.stack.append(mode)

    def pop(self):
        return self.stack.pop()

    @property
    def is_read(self):
        return self.stack[-1] == IOMode.READ

    @property
    def is_write(self):
        return self.stack[-1] == IOMode.WRITE


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
    }
    for _, set_def in model_obj.sets.items():
        metrics["n_sets"] += 1
        logger.debug("collecting inouts for set %s", set_def.name)
        patch_model(visitor)
        for _, instr_def in set_def.instructions.items():
            metrics["n_instructions"] += 1
            context = VisitorContext()
            logger.debug("collecting inouts for instr %s", instr_def.name)
            try:
                instr_def.operation.generate(context)
                for op_name, op_def in instr_def.operands.items():
                    if op_name in context.reads and op_name in context.writes:
                        if seal5.model.Seal5OperandAttribute.INOUT not in instr_def.attributes:
                            op_def.attributes[seal5.model.Seal5OperandAttribute.INOUT] = []
                    elif op_name in context.reads:
                        if seal5.model.Seal5OperandAttribute.IN not in instr_def.attributes:
                            op_def.attributes[seal5.model.Seal5OperandAttribute.IN] = []
                    elif op_name in context.writes:
                        if seal5.model.Seal5OperandAttribute.OUT not in instr_def.attributes:
                            op_def.attributes[seal5.model.Seal5OperandAttribute.OUT] = []
                # print("---")
                # print("instr_def.scalars.keys()", instr_def.scalars.keys())
                for reg_name in context.reads:
                    if reg_name == "PC":
                        continue
                    # print("reg_name1", reg_name)
                    if reg_name in instr_def.operands.keys():
                        continue
                    if reg_name in instr_def.scalars.keys():
                        continue
                    # TODO: how about register groups
                    # TODO: handle other architectural state vars here (PC,...)
                    assert reg_name in set_def.registers
                    uses = instr_def.attributes.get(seal5.model.Seal5InstrAttribute.USES, [])
                    uses.append(reg_name)
                    # TODO: drop duplicates?
                    instr_def.attributes[seal5.model.Seal5InstrAttribute.USES] = uses
                for reg_name in context.writes:
                    # print("reg_name2", reg_name)
                    if reg_name == "PC":
                        continue
                    if reg_name in instr_def.operands.keys():
                        continue
                    if reg_name in instr_def.scalars.keys():
                        continue
                    # TODO: how about register groups
                    # TODO: handle other architectural state vars here (PC,...)
                    assert reg_name in set_def.registers
                    defs = instr_def.attributes.get(seal5.model.Seal5InstrAttribute.DEFS, [])
                    defs.append(reg_name)
                    # TODO: drop duplicates?
                    instr_def.attributes[seal5.model.Seal5InstrAttribute.DEFS] = defs
                # print("instr_def.operands_", instr_def.operands)
                # print("instr_def.attributes", instr_def.attributes)
                # input("999")
                metrics["n_success"] += 1
                metrics["success_instructions"].append(instr_def.name)
            except Exception as ex:
                logger.exception(ex)
                metrics["n_failed"] += 1
                metrics["failed_instructions"].append(instr_def.name)

    dump_model(model_obj, out_path, compat=args.compat)
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Patch utils for seal5."""
import os
import argparse
from pathlib import Path
from typing import Optional

from email.utils import formatdate
import yaml

from seal5 import utils
from seal5.logging import get_logger

logger = get_logger()


def generate_patch(index_file, llvm_dir=None, out_file=None, author=None, mail=None, msg=None, append=None):
    """Generate patch contents based on index file."""
    del append  # unused
    # base_dir = os.path.dirname(index_file)
    if msg:
        if not isinstance(msg, str):
            assert isinstance(msg, list)
            assert len(msg) == 1
            msg = msg[0]
    else:
        msg = "Placeholder"

    with open(index_file, "r", encoding="utf-8") as file:
        index = yaml.safe_load(file)

    global_artifacts = index["artifacts"]

    def generate_patch_set(fragments):
        ps = ""
        # Build minimal patch header for 'git am'
        user_name = author
        user_email = mail
        ps += f"From: {user_name} <{user_email}>\n"
        ps += f"Date: {formatdate()}\n"
        ps += f"Subject: {msg}\n\n\n"
        for fragment in fragments:
            ps += fragment
        return ps

    def find_site(path, key):
        # print("find_site", path, key)
        # start_mark = key + ' - INSERTION_START'
        # enc_mark = key + ' - INSERTION_END'
        if key:
            start_mark_ = path.split("/")[-1] + " - " + key + " - INSERTION_START"
            end_mark_ = path.split("/")[-1] + " - " + key + " - INSERTION_END"
        else:
            start_mark_ = path.split("/")[-1] + " - INSERTION_START"
            end_mark_ = path.split("/")[-1] + " - INSERTION_END"
        # print("start", start_mark_)
        # print("end", end_mark_)
        site_line = -1
        site_len = 0
        start_mark = None
        fullpath = path if llvm_dir is None else os.path.join(llvm_dir, path)
        with open(fullpath, "r", encoding="utf-8") as src:
            lines = src.readlines()
            for line in lines:
                if line.find(start_mark_) != -1:
                    start_mark = line
                    site_line = lines.index(line)
                elif line.find(end_mark_) != -1:
                    end_mark = line.rstrip("\n")
                    site_len = lines.index(line) - site_line + 1
                    break
                elif site_line >= 0:
                    # Accumulate all lines between the start mark and the end
                    # as part of the mark, so the new lines get injected just
                    # before the end mark
                    start_mark += line
        if start_mark is None:
            # fallback
            logger.warning("Marker not found: {path}, {key}. Retrying without key...")
            if key:
                return find_site(path, None)
            assert False, f"Marker not found: {path}, {key}"
        return site_line, site_len, start_mark, end_mark

    def generate_patch_fragment(artifact):
        dest_path = artifact.get("dest_path", None)
        src_path = artifact.get("src_path", None)
        key = artifact.get("key", None)
        start = artifact.get("start", None)
        end = artifact.get("end", None)
        line = artifact.get("line", None)
        content_ = artifact.get("content", None)
        assert dest_path is not None
        is_file = False
        is_dir = False
        is_patch = False
        # TODO: do not depend on dest_path
        if key:  # NamedPatch
            is_patch = True
        elif start is not None and end is not None:
            raise NotImplementedError
        elif line is not None:
            raise NotImplementedError
        else:  # File/Dir
            if content_ is None:
                if Path(src_path).is_file():
                    is_file = True
                elif Path(src_path).is_dir():
                    is_dir = True
                else:
                    raise RuntimeError(f"File not found: {src_path}")
            else:
                is_file = True
        if content_:
            assert isinstance(content_, str)
        else:
            if is_patch or is_file:
                with open(src_path, "r", encoding="utf-8") as f:
                    content_ = f.read()
            elif is_dir:
                files = Path(src_path).rglob("*")
                base = dest_path
                # print("base", base)
                # print("files", files)
                fragments = []
                for file in files:
                    if file.is_dir():
                        continue
                    # print("file", file)
                    dest_path_ = str(Path(base) / (str(file).replace(f"{src_path}/", "")))
                    # print("dest_path_", dest_path_)
                    src_path_ = file
                    file_artifact = {
                        "dest_path": str(dest_path_),
                        "src_path": str(src_path_),
                    }
                    fragment = generate_patch_fragment(file_artifact)
                    fragments.append(fragment)
                # print("fragments", fragments)
                # input("AAA")
                return "".join(fragments)
                # raise NotImplementedError
            else:
                assert False
        content = "+" + content_.replace("\n", "\n+")
        if is_patch:
            # Updating existing file
            orig_file = "a/" + dest_path
            new_file = "b/" + dest_path
            site_line, site_len, start_mark, end_mark = find_site(dest_path, key)
            new_start = site_line + 1
            new_len = content_.count("\n") + 1 + site_len
            # ensure all existing lines in the match prefixed by a space
            # if site_len > 2 and start_mark[-1] == "\n":
            # if False:
            #     start_mark = start_mark.rstrip("\n").replace("\n", "\n ")
            # else:
            start_mark = start_mark.rstrip("\n").replace("\n", "\n ")
            content = f" {start_mark}\n{content}\n {end_mark}"
        else:
            assert is_file
            # Adding new file
            orig_file = "/dev/null"
            new_file = "b/" + dest_path
            new_start = 1
            site_len = 0
            site_line = -1
            new_len = content_.count("\n") + 1
        return f"""--- {orig_file}
+++ {new_file}
@@ -{site_line + 1},{site_len} +{new_start},{new_len} @@
{content}
"""

    fragments = []

    def process_artifacts(artifacts):
        ret = []
        for artifact in artifacts:
            fragment = generate_patch_fragment(artifact)
            ret.append(fragment)
        return ret

    fragments.extend(process_artifacts(global_artifacts))

    for extn in index["extensions"]:
        fragments.extend(process_artifacts(extn["artifacts"]))

    patch_set = generate_patch_set(fragments)

    if out_file:
        with open(out_file, "w", encoding="utf-8") as f:
            f.write(patch_set)
    else:
        print(patch_set)


def process_arguments():
    """Progress cmdline arguments."""
    parser = argparse.ArgumentParser(description="Generate patch file for adding extension to LLVM")
    parser.add_argument("index_file", help="list of code fragments to inject")
    parser.add_argument("--llvm-dir", type=str, default=None, help="path to llvm repo (default: cwd)")
    parser.add_argument("--out-file", type=str, default=None, help="write patchset to file (default: print to stdout)")
    parser.add_argument("--author", type=str, default="bob", help="author name in patchset")
    parser.add_argument("--mail", type=str, default="bob@bob.com", help="author mail in patchset")
    parser.add_argument(
        "--msg", type=str, nargs="+", default="[PATCH] Instructions injected", help="custom text for commit message"
    )
    parser.add_argument(
        "--append",
        "-a",
        action="store_true",
        help="whether existing patches should be appended instead of beeing overwritten",
    )
    args = parser.parse_args()
    return args


def analyze_diff(repo, base: str, cur: Optional[str] = None):
    args = ["git", "diff", "--shortstat"]
    args += [base]
    if cur is not None:
        args += [cur]
    out = utils.exec_getout(
        *args,
        cwd=repo.working_tree_dir,
        print_func=lambda *args, **kwargs: None,
        live=False,
    )
    n_files_changed = 0
    n_insertions = 0
    n_deletions = 0
    for x in out.split(","):
        x = x.strip()
        if len(x) == 0:
            continue
        val, key = x.split(" ", 1)
        val = int(val)
        if "files changed" in key or "file changed" in key:
            n_files_changed = val
        elif "insertions" in key:
            n_insertions = val
        elif "deletions" in key:
            n_deletions = val
    return n_files_changed, n_insertions, n_deletions


def main():
    """Main entry point."""
    args = process_arguments()
    generate_patch(
        args.index_file,
        llvm_dir=args.llvm_dir,
        out_file=args.out_file,
        author=args.authot,
        mail=args.mail,
        msg=args.msg,
        append=args.append,
    )


if __name__ == "__main__":
    main()



from typing import Optional

from seal5.settings import RISCVSettings


def get_riscv_defaults(riscv_settings: Optional[RISCVSettings] = None):
    default_features = ["+m", "+fast-unaligned-access"]
    default_xlen = None
    if riscv_settings:
        default_features_ = riscv_settings.features
        if default_features_ is not None:
            default_features = default_features_
        default_xlen_ = riscv_settings.xlen
        if default_xlen is not None:
            default_xlen = default_xlen_
    return default_features, default_xlen


def build_riscv_mattr(features, xlen: Optional[int] = None):
    mattrs = []

    def fix_prefix(x):
        assert len(x) > 0
        if x[0] in ["-", "+"]:
            return x
        return f"+{x}"

    if features is not None:
        mattrs += list(features)
    if xlen is not None:
        if xlen == 64:
            mattrs += ["64bit"]
    mattrs = list(set(map(fix_prefix, mattrs)))
    mattr = ",".join(mattrs)
    return mattr



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Loging utilities for Seal5."""

import logging
import logging.handlers
import sys

logging.basicConfig(
    level=logging.INFO,
    format="[%(asctime)s]::%(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

INITIALIZED = False


def get_formatter(minimal=False):
    """Returns a log formatter for one on two predefined formats."""
    if minimal:
        fmt = "%(levelname)s - %(message)s"
    else:
        fmt = "[%(asctime)s]::%(pathname)s:%(lineno)d::%(levelname)s - %(message)s"
    formatter = logging.Formatter(fmt)
    return formatter


def get_logger():
    """Helper function which return the main seal5 logger while ensuring that is is properly initialized."""
    global INITIALIZED
    # root_logger = logging.getLogger()
    # root_logger.setLevel(logging.DEBUG)
    # root_logger.setLevel(logging.INFO)
    logger = logging.getLogger("seal5")
    # logger.setLevel(logging.DEBUG)
    if len(logger.handlers) == 0:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(get_formatter(minimal=True))
        # stream_handler.setLevel(logging.DEBUG)
        logger.addHandler(stream_handler)
        logger.propagate = False
        INITIALIZED = True
    return logger


def set_log_level(console_level=None, file_level=None):
    """Set command line log level at runtime."""
    # print("set_log_level", console_level, file_level)
    logger = logging.getLogger("seal5")
    for handler in logger.handlers[:]:
        # print("handler", handler, type(handler))
        if (
            isinstance(handler, (logging.FileHandler, logging.handlers.TimedRotatingFileHandler))
            and file_level is not None
        ):
            handler.setLevel(file_level)
            # print("NEWIF")
        elif isinstance(handler, logging.StreamHandler) and console_level is not None:
            handler.setLevel(console_level)
            # print("NEWELIF")
    logger.setLevel(logging.DEBUG)


def set_log_file(path, level=logging.DEBUG, rotate=False):
    """Enable logging to a file."""
    # print("set_log_file", level)
    del level  # TODO: use level
    logger = logging.getLogger("seal5")
    logger.setLevel(logging.DEBUG)
    if rotate:
        file_handler = logging.handlers.TimedRotatingFileHandler(filename=path, when="midnight", backupCount=30)
    else:
        file_handler = logging.FileHandler(path, mode="a")
    file_handler.setFormatter(get_formatter())
    # file_handler.setLevel(level)
    file_handler.setLevel(logging.DEBUG)
    for handler in logger.handlers[:]:
        if isinstance(handler, logging.FileHandler):
            logger.removeHandler(handler)
    logger.addHandler(file_handler)



#!/usr/bin/env python
#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""The setup script."""

import os
from setuptools import setup, find_packages
from seal5.version import __version__


def resource_files(directory):
    paths = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


with open("README.md") as readme_file:
    readme = readme_file.read()


def get_requirements():
    return []


requirements = get_requirements()

setup(
    author="TUM Department of Electrical and Computer Engineering - Chair of Electronic Design Automation",
    author_email="philipp.van-kempen@tum.de",
    python_requires=">=3.8",
    classifiers=[  # TODO
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    description="Seal5 - Semi-automated LLVM Support for RISC-V Extensions including Autovectorization",
    entry_points={
        "console_scripts": [
            "seal5=seal5.cli.main:main",
        ],
    },
    install_requires=requirements,
    extras_require={},
    license="Apache License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="seal5",
    name="seal5",
    packages=find_packages(include=["seal5", "seal5.*"]),
    package_data={"seal5": resource_files("resources")},
    test_suite="tests",
    tests_require=requirements,
    url="https://github.com/tum-ei-eda/seal5",
    version=__version__,
    zip_safe=False,
)
